from enum import Enum
import sys, os

class Colours(Enum):
    """Enum class containing codes for most console colours. (Added in v2.0 rw)"""
    GREEN = "\u001b[32m"
    BLUE = "\u001b[36m"
    CLAY = "\u001b[38;5;51m"
    PURPLE = "\u001b[38;5;200m"
    RED = "\u001b[31m"
    BOLD_RED = "\u001b[31;1m"
    PINK_GREY = "\u001b[38;5;139m"
    GREY = "\u001b[38;20m"
    YELLOW = "\u001b[33;20m"
    ORANGE = "\u001b[38;5;214m"

    RESET_COLOUR = "\u001b[0m"

    def __init__(self, colour_value:str):
        if sys.platform == "win32": os.system("color")

    def apply_to_string(self, string:str) -> str:
        """Returns that string but with this colour applied to it."""
        return self.value + string + self.RESET_COLOUR.value


import logging as log

from .colours import Colours

class DevGoldyUtilsError(Exception):
    """Raises whenever there's a known error in devgoldyutils."""
    def __init__(self, message:str, logger:log.Logger = None):
        message = Colours.RED.apply_to_string(message)

        if logger is not None:
            logger.error(message)
        
        super().__init__(message)


from __future__ import annotations
from typing import Any, Dict

import logging as log
from dataclasses import dataclass, field

from . import errors

@dataclass
class DictDataclass:
    """A base dataclass for handling dict filled dataclasses. I really don't know what to call it but I inherit from this to obtain special methods."""
    logger:log.Logger = field(init=False, repr=False, default=None)
    """A logger for us to use to send warnings about key errors."""

    def __post_init__(self):
        super().__post_init__(self)

    def get(self, *keys, data = None, default_value = None) -> Any|Dict:
        """
        A small method used to grab data from the ``data`` dictionary with an advantage of handling KeyError respectfully. 
        It's better to use this method instead of just directly accessing the dictionary.
        ------------------

        Parameters
        ----------
        ``*keys``
            The keys to get to the dictionary object. E.g ``self.data['cities']['london']['population']`` would be written like --> ``self.get('cities', 'london', 'population')``.

        ``data``
            The dictionary containing the data, if left none we'll use ``self.data``.

        ``default_value``
            The default value to return if the object is not found in the dictionary.

        Returns
        -------
        ``Any`` | ``Dict``
            Can return anything from the dictionary.
        """
        if data is None:
            try:
                data = self.data
            except AttributeError as e:
                raise errors.DevGoldyUtilsError(
                    f"Looks like the dataclass '{self.__class__.__name__}' doesn't countian a data dict field. Please add one or use the data parameter in self.get() instead. \nError --> {e}"
                )

        try:
            for key in keys:
                data = data[key]
            
            return data
        except (KeyError, TypeError) as e:

            if self.logger is not None:
                self.logger.warning(f"Could not find key {e} in dict so I'm returning default value '{default_value}'... Keys: {keys}")

            return default_value


import random

class LegacyColours:
    """Class that contants methods for colouring the text printed to the console."""

    def RANDOM(self, text:str):
        random_num = random.randint(1, 4)

        if random_num == 1: return self.GREEN(text)
        if random_num == 2: return self.BLUE(text)
        if random_num == 3: return self.CLAY(text)
        if random_num == 4: return self.PURPLE(text)

    def PICK_COLOUR(self, colour_name:str, text:str):
        if not colour_name == None:
            if colour_name.lower() in ["green"]:
                return self.GREEN(text)

            if colour_name.lower() in ["blue"]:
                return self.BLUE(text)

            if colour_name.lower() in ["clay"]:
                return self.CLAY(text)

            if colour_name.lower() in ["purple"]:
                return self.PURPLE(text)
        else:
            return self.RANDOM(text)
    
    def GREEN(self, text:str):
        return "\u001b[32m" + str(text) + "\u001b[0m"

    def BLUE(self, text:str):
        return "\u001b[36m" + str(text) + "\u001b[0m"

    def CLAY(self, text:str):
        return "\u001b[38;5;51m" + str(text) + "\u001b[0m"

    def PURPLE(self, text:str):
        return "\u001b[38;5;200m" + str(text) + "\u001b[0m"

    def RED(self, text:str):
        return "\u001b[31m" + str(text) + "\u001b[0m"


import logging as log
from .colours import Colours

# Logging module stuff
# -----------------------
class CustomFormatter(log.Formatter):
    """This is an internal thing but it creates the amazing logging format I use in all my python programs."""
    format_string = "[%(levelname)s]\u001b[0m (%(name)s) - %(message)s"

    FORMATS = {
        log.DEBUG: Colours.PINK_GREY.value + format_string,
        log.INFO: Colours.CLAY.value + format_string,
        log.WARNING: Colours.YELLOW.value + format_string,
        log.ERROR: Colours.RED.value + format_string,
        log.CRITICAL: Colours.BOLD_RED.value + format_string
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = log.Formatter(log_fmt)
        return formatter.format(record)

class LoggerAdapter(log.LoggerAdapter):
    """Allows you to create and bind another logger to a current logger."""
    def __init__(self, logger:log.Logger, prefix:str):
        super().__init__(logger, {'prefix': prefix})

    def process(self, msg, kwargs):
        return f"\u001b[92m[{self.extra['prefix']}\u001b[92m]\u001b[0m {msg}", kwargs

# Method for adding custom handler to any logger object.
# -------------------------------------
def add_custom_handler(logger:log.Logger) -> log.Logger:
    """Method for adding custom handler to any logger object."""
    stream_handler = log.StreamHandler()
    stream_handler.setLevel(log.DEBUG)
    stream_handler.setFormatter(CustomFormatter())

    logger.propagate = False
    logger.addHandler(stream_handler)

    return logger


from dataclasses import dataclass, field
from logging import getLogger
from devgoldyutils import DictDataclass

@dataclass
class Test(DictDataclass):
    data:dict
    bruh:str = field(init=False)

    logger = getLogger("bruhhhh")

    def __post_init__(self):
        self.bruh = self.get("ow")

test = Test({
    "owo": {
        "bruh": "omg"
    }
})

print(test.bruh)


# V1 support. (See 'https://gist.github.com/THEGOLDENPRO/34dccf77a3c0eb2d86bfebe90502e4c4' for more info.)
# -----------------------------------------------------------------------------------------------------------
import sys
from .legacy_colours import LegacyColours
setattr(sys.modules[__name__], "Colours", LegacyColours)


import random
import os, sys

from . import colours

class LegacyConsole(colours.LegacyColours):
    def __init__(self):
        if sys.platform == "win32": os.system("color")

# V1 support. (See 'https://gist.github.com/THEGOLDENPRO/34dccf77a3c0eb2d86bfebe90502e4c4' for more info.)
setattr(sys.modules[__name__], "Console", LegacyConsole)